"""
Tests for Kernel Shap Module
"""
import numpy as np
import tensorflow as tf

from xplique.attributions import KernelShap
from ..utils import generate_data, generate_model, almost_equal

def test_get_probs():
    """
    Ensure the get_probs function output has the right shape and return
    right values for simple example
    """
    num_features = 5
    probs = KernelShap._get_probs_nb_selected_feature(num_features)
    expected_probs = np.array([0.,1.,2/3,2/3,1.])
    assert almost_equal(probs.numpy(),expected_probs, epsilon=1e-5)
    num_features = [4,6,9,7]
    for num_feature in num_features:
        probs = KernelShap._get_probs_nb_selected_feature(num_feature)
        assert probs.shape == num_feature

def test_pertub_func():
    """
    Ensure the pertub function output has the right shape and
    values belong to (0, num_interp_features)
    """
    interpret_samples = KernelShap._kernel_shap_pertub_func(10,20)
    assert interpret_samples.shape == (20,10)
    for sample in interpret_samples:
        assert tf.less(tf.reduce_sum(sample, axis=0),10)
        assert tf.greater(tf.reduce_sum(sample, axis=0),0)

def test_output_explain():
    """The output shape must be the same as the input shape, except for the channels"""
    input_shapes = [(28, 28, 1), (32, 32, 3)]
    nb_labels = 10

    def map_four_by_four(inp):

        width = inp.shape[0]
        height = inp.shape[1]

        mapping = np.zeros((width,height))
        for i in range(width):
            if i%2 != 0:
                mapping[i] = mapping[i-1]
            else:
                for j in range(height):
                    mapping[i][j] = (width/2) * (i//2) + (j//2)

        mapping = tf.cast(mapping, dtype=tf.int32)
        return mapping

    for input_shape in input_shapes:
        samples, labels = generate_data(input_shape, nb_labels, 20)
        model = generate_model(input_shape, nb_labels)

        method = KernelShap(model, map_to_interpret_space=map_four_by_four, nb_samples=10)

        explanations = method.explain(samples, labels)
        assert samples.shape[:3] == explanations.shape

def test_inputs_batching():
    """ Ensure, that we can call explain with batched inputs """
    nb_labels = 10

    samples, labels = generate_data((32, 32, 3), nb_labels, 200)
    model = generate_model((32, 32, 3), nb_labels)

    method = KernelShap(model,
                    batch_size=15,
                    nb_samples=20)

    explanations = method.explain(samples, labels)

    assert samples.shape[:3] == explanations.shape



"""
Stability (or Sensitivity) metrics
"""

import numpy as np
import tensorflow as tf

from .base import ExplainerMetric
from ..types import Callable, Optional, Union


class AverageStability(ExplainerMetric):
    """
    Used to compute the average sensitivity metric (or stability). This metric ensure that close
    inputs with similar predictions yields similar explanations. For each inputs we randomly
    sample noise to add to the inputs and compute the explanation for the noisy inputs. We then
    get the average distance between the original explanations and the noisy explanations.

    Ref. Bhatt & al., Evaluating and Aggregating Feature-based Model Explanations (2020).
    https://arxiv.org/abs/2005.00631 (def. 2)

    Parameters
    ----------
    model
        Model used for computing metric.
    inputs
        Input samples under study.
    targets
        One-hot encoded labels or regression target (e.g {+1, -1}), one for each sample.
    batch_size
        Number of samples to explain at once, if None compute all at once.
    radius
        Maximum value of the uniform noise added to the inputs before recalculating their
        explanations.
    distance
        Distance metric between the explanations.
    nb_samples
        Number of different neighbors points to try on each input to measure the stability.
    """

    def __init__(self,
                 model: Callable,
                 inputs: Union[tf.data.Dataset, tf.Tensor, np.ndarray],
                 targets: Optional[Union[tf.Tensor, np.ndarray]] = None,
                 batch_size: Optional[int] = 64,
                 radius: float = 0.1,
                 distance: Union[str, Callable] = 'l2',
                 nb_samples: int = 20):
        # pylint: disable=R0913
        super().__init__(model, inputs, targets, batch_size)
        self.radius = radius
        self.nb_samples = nb_samples

        if distance == 'l1':
            self.distance = lambda x, y: tf.reduce_sum(tf.abs(x - y))
        elif distance == 'l2':
            self.distance = lambda x, y: tf.reduce_sum((x-y)**2.0)
        elif hasattr(distance, '__call__'):
            self.distance = distance
        else:
            raise ValueError(f"{distance} is not a valid distance.")

        # prepare the noisy masks that will be used to generate the neighbors
        self.noisy_masks = tf.random.uniform((nb_samples, *inputs.shape[1:]), 0,
                                             self.radius)

    def evaluate(self,
                 explainer: Callable,
                 base_explanations: Optional[Union[tf.Tensor, np.ndarray]] = None,
        ) -> float:
        # pylint: disable=W0221
        """
        Evaluate the fidelity score.

        Parameters
        ----------
        explainer
            Explainer or Explanations associated to each inputs.
        base_explanations
            Explanation for the inputs under study. Calculates them automatically if they are
            not provided.

        Returns
        -------
        stability_score
            Average distance between the explanations
        """
        if base_explanations is None:
            base_explanations = np.array(explainer(self.inputs, self.targets))

        distances = []
        for inp, label, phi in zip(self.inputs, self.targets, base_explanations):
            label = tf.repeat(label[None, :], self.nb_samples, 0)
            neighbors = inp + self.noisy_masks
            phis_neighbors = np.array(explainer(neighbors, label))

            # compute the distances between each new explanations
            avg_dist = np.mean([self.distance(phi_n, phi) for phi_n in phis_neighbors])
            distances.append(avg_dist)

        stability_score = np.mean(distances)

        return float(stability_score)



import tensorflow as tf
import numpy as np

from xplique.features_visualizations import Objective
from ..utils import almost_equal


def dummy_model():
    model = tf.keras.Sequential([
        tf.keras.layers.Input((28, 28, 3)),
        tf.keras.layers.Conv2D(16, (3, 3)),
        tf.keras.layers.Conv2D(16, (3, 3), name="early"),
        tf.keras.layers.MaxPool2D((2, 2)),
        tf.keras.layers.Conv2D(16, (3, 3)),
        tf.keras.layers.Conv2D(16, (3, 3), name="features"),
        tf.keras.layers.MaxPool2D((2, 2)),
        tf.keras.layers.Flatten(name="pre-logits"),
        tf.keras.layers.Dense(10, name="logits")
    ])
    model.compile()
    return model


def test_layer():
    """ Ensure we can target a layer """
    model = dummy_model()

    objectif = Objective.layer(model, -1)

    mask = np.ones(model.get_layer("logits").output.shape[1:])
    assert almost_equal(objectif.masks[0], mask)

    model_reconf, objective_func, names, input_shape = objectif.compile()

    assert model_reconf.outputs[0].name == model.layers[-1].output.name
    assert names[0] == "Layer#logits"
    assert input_shape == (1, 28, 28, 3)


def test_direction():
    """ Ensure we can target a direction """
    model = dummy_model()

    direction_vec = np.random.random(model.get_layer("early").output.shape[1:])
    objectif = Objective.direction(model, "early", direction_vec)

    assert almost_equal(objectif.masks[0], direction_vec)

    model_reconf, objective_func, names, input_shape = objectif.compile()

    assert model_reconf.outputs[0].name == model.layers[1].output.name
    assert names[0] == "Direction#early_0"
    assert input_shape == (1, 28, 28, 3)


def test_channel():
    """ Ensure we can target a channel """
    model = dummy_model()
    nb_channels = 10 # number of channel to optim (1...n)

    objectif = Objective.channel(model, "early", list(range(nb_channels)))

    # the first mask should target the first channel
    mask_1 = np.zeros(model.layers[1].output.shape[1:])
    mask_1[:, :, 0] = 1.0
    assert almost_equal(objectif.masks[0][0], mask_1)

    model_reconf, objective_func, names, input_shape = objectif.compile()

    # ensure we are targeting the good layer (target by id should be equal to target by name)
    assert model_reconf.outputs[0].name == model.layers[1].output.name
    assert all([names[i] == f"Channel#early_{i}" for i in range(nb_channels)])
    # we should have one input for each optim (one for each of the channels
    assert input_shape == (10, 28, 28, 3)


def test_neurons():
    """ Ensure we can target neurons """
    model = dummy_model()

    objectif = Objective.neuron(model, "early", list(range(20)))

    masks = np.zeros((20) + model.get_layer("early").output.shape[1:])
    for i in range(20):
        masks[i, 0, i // masks.shape[3], i % masks.shape[3]] = 1.0

    assert almost_equal(objectif.masks[0], masks)

    model_reconf, objective_func, names, input_shape = objectif.compile()

    assert model_reconf.outputs[0].name == model.layers[1].output.name
    assert all([names[i] == f"Neuron#early_{i}" for i in range(20)])
    assert input_shape == (20, 28, 28, 3)


def test_combinations():
    """ Ensure the combinations of the objective are correct """
    model = dummy_model()

    layer_obj = Objective.layer(model, -1)

    direction_vec = np.random.random(model.get_layer("pre-logits").output.shape[1:])
    direction_obj = Objective.direction(model, "pre-logits", direction_vec)

    nb_channels = 5
    channels_obj = Objective.channel(model, "early", list(range(nb_channels)))
    mask_channel_0 = np.zeros(model.get_layer("early").output.shape[1:])
    mask_channel_0[:, :, 0] = 1.0

    nb_neurons = 3
    neurons_obj = Objective.neuron(model, "features", list(range(nb_neurons)))
    mask_neurons = np.zeros((nb_neurons) + model.get_layer("features").output.shape[1:])
    for i in range(nb_neurons):
        mask_neurons[i, 0, i // mask_neurons.shape[3], i % mask_neurons.shape[3]] = 1.0

    combined_obj = layer_obj + direction_obj + channels_obj + neurons_obj

    # the generation of masks must remain similar to the operation on singular objectives
    assert len(combined_obj.masks) == 4
    assert almost_equal(combined_obj.masks[0][0], np.ones(model.layers[-1].output.shape[1:]))
    assert almost_equal(combined_obj.masks[1][0], direction_vec)
    assert almost_equal(combined_obj.masks[2][0], mask_channel_0)
    assert almost_equal(combined_obj.masks[3], mask_neurons)

    model_reconf, objective_func, names, input_shape = combined_obj.compile()

    assert [o.name for o in model_reconf.outputs] == [model.layers[-1].output.name, # layer obj
                                                      model.layers[-2].output.name, # direction obj
                                                      model.layers[1].output.name,  # channel obj
                                                      model.layers[-4].output.name] # neuron obj
    assert_names = [
        f"Layer#logits & Direction#pre-logits_0 & Channel#early_{c_id} & Neuron#features_{n_id}" for
        c_id in range(nb_channels) for n_id in range(nb_neurons)]
    assert all([names[i] == assert_names[i] for i in range(nb_neurons*nb_channels)])
    assert input_shape == (nb_neurons * nb_channels, 28, 28, 3) # number of combinations



import tensorflow as tf
import numpy as np

from ..utils import (generate_model, generate_regression_model, generate_timeseries_model, 
                     generate_data, almost_equal)
from xplique.commons.operators import regression_operator
from xplique.metrics import Insertion, Deletion, MuFidelity


def test_mu_fidelity():
    # ensure we can compute the metric with consistents arguments
    input_shape, nb_labels, nb_samples = ((32, 32, 3), 10, 20)
    x, y = generate_data(input_shape, nb_labels, nb_samples)
    model = generate_model(input_shape, nb_labels)
    explanations = np.random.uniform(0, 1, x.shape[:-1])

    nb_estimation = 10 # number of samples to test correlation for each samples

    for grid_size in [None, 5]:
        for subset_percent in [0.1, 0.9]:
            for baseline_mode in [0.0, lambda x : x-0.5]:
                score = MuFidelity(model, x, y, grid_size=grid_size,
                                   subset_percent=subset_percent,
                                   baseline_mode=baseline_mode,
                                   nb_samples=nb_estimation)(explanations)
                assert -1.0 < score < 1.0


def test_causal_metrics():
    # ensure we can compute insertion/deletion metric with consistent arguments
    input_shape, nb_labels, nb_samples = ((32, 32, 3), 10, 20)
    x, y = generate_data(input_shape, nb_labels, nb_samples)
    model = generate_model(input_shape, nb_labels)
    explanations = np.random.uniform(0, 1, x.shape[:-1])

    for step in [5, 10]:
        for baseline_mode in [0.0, lambda x: x-0.5]:
            score_insertion = Insertion(model, x, y,
                                        baseline_mode=baseline_mode,
                                        steps=step)(explanations)
            score_deletion = Deletion(model, x, y,
                                      baseline_mode=baseline_mode,
                                      steps=step)(explanations)

            for score in [score_insertion, score_deletion]:
                assert 0.0 <= score <= 1.0


def test_data_types_with_regression():
    # ensure that the different metrics are robust to several data types with regression
    # test several data types
    data_types = {
        "tabular": (16,),
        "time-series": (16, 8),
        "images": (16, 16, 3)
    }

    for type, input_shape in data_types.items():
        nb_labels, samples = 10, 20
        x, y = generate_data(input_shape, nb_labels, samples)
        model = generate_regression_model(input_shape, nb_labels)

        explanations = np.random.uniform(0, 1, x.shape[:-1])

    score_insertion = Insertion(
        model, x, y, operator=regression_operator, steps=3,
    ).evaluate(explanations)

    score_deletion = Deletion(
        model, x, y, operator=regression_operator, steps=3,
    ).evaluate(explanations)

    for score in [score_insertion, score_deletion]:
        assert 0.0 < score

    score_mufidelity = MuFidelity(
        model, x, y, operator=regression_operator, nb_samples=3
    ).evaluate(explanations)
            
    assert -1.0 < score_mufidelity < 1.0


def test_perfect_correlation():
    """Ensure we get perfect score if the correlation is perfect"""
    # we ensure perfect correlation if the model return the sum of the input,
    # and the input is the explanations: corr( sum(phi), sum(x) - sum(x-phi) )
    # to do so we define f(x) -> sum(x) and phi = x
    nb_classes = 2

    input_shape, nb_labels, nb_samples = ((32, 32, 1), nb_classes, 20)
    x, y = generate_data(input_shape, nb_labels, nb_samples)
    model = lambda x: tf.repeat(tf.reduce_sum(x, axis=(1, 2, 3))[:, None], nb_classes, -1)
    explanations = x

    perfect_score = MuFidelity(model, x, y, grid_size=None,
                               subset_percent=0.1,
                               baseline_mode=0.0,
                               nb_samples=200)(explanations)
    assert almost_equal(perfect_score, 1.0)


def test_worst_correlation():
    """Ensure we get worst score if the correlation is inversed"""
    # we ensure worst correlation if the model return the -sum of the input,
    # and the input is the explanations: corr( sum(phi), sum(x) - sum(x-phi) )
    # to do so we define f(x) -> sum(x) and phi = x
    nb_classes = 2

    input_shape, nb_labels, nb_samples = ((32, 32, 1), nb_classes, 20)
    x, y = generate_data(input_shape, nb_labels, nb_samples)
    model = lambda x: tf.repeat(tf.reduce_sum(x, axis=(1, 2, 3))[:, None], nb_classes, -1)
    explanations = -x

    perfect_score = MuFidelity(model, x, y, grid_size=None,
                               subset_percent=0.1,
                               baseline_mode=0.0,
                               nb_samples=200)(explanations)
    assert almost_equal(perfect_score, -1.0)


def test_perfect_deletion():
    """Ensure we get perfect score if the model is sensible to deletion"""
    # we ensure perfect deletion if the model return 0.0 as soon as there is
    # one element set to baseline
    dim = 16
    steps = dim**2

    input_shape, nb_labels, nb_samples = ((dim, dim, 1), 2, 20)
    x, y = generate_data(input_shape, nb_labels, nb_samples)

    model = lambda x: 1.0 - tf.reduce_max(tf.cast(x == 0.0, tf.float32), (1, 2))
    explanations = x

    perfect_score = Deletion(model, x, y, steps=steps)(explanations)
    assert almost_equal(perfect_score, 0.0, 1e-2)


def test_perfect_insertion():
    """Ensure we get perfect score if the model is sensible to insertion"""
    # we ensure perfect deletion if the model return 1.0 as soon as there is
    # one element to non-baseline
    dim = 16
    steps = dim**2

    input_shape, nb_labels, nb_samples = ((dim, dim, 1), 2, 20)
    x, y = generate_data(input_shape, nb_labels, nb_samples)

    model = lambda x: tf.reduce_max(tf.cast(x != 0.0, tf.float32), (1, 2))
    explanations = x

    perfect_score = Insertion(model, x, y, steps=steps)(explanations)
    assert almost_equal(perfect_score, 1.0, 1e-2)



"""
Xplique
-------

The goal of Xplique is to provide a simple interface to the latest explanation
techniques
"""

__version__ = '1.2.1'

from . import attributions
from . import concepts
from . import features_visualizations
from . import commons
from . import plots

from .commons import Tasks



"""
Kernel functions for HSIC estimators.
"""
# pylint: disable=C0103

from enum import Enum
import tensorflow as tf


@tf.function
def rbf(X, Y, width):
    """
    Radial Basis Function kernel.

    Parameters
    ----------
    X
        Samples of input variable
    Y
        Samples of output variable

    Returns
    -------
    Kernel matrix
    """
    XY = (X - Y) ** 2
    K = tf.exp(-XY / (2 * width**2))
    return K


@tf.function
def binary(X, Y):
    """
    Dirac kernel

    Parameters
    ----------
    X
        Samples of input variable (must be binary)
    Y
        Samples of output variable

    Returns
    -------
    Kernel matrix
    """
    K = 0.5 - (X - Y) ** 2
    return K


@tf.function
def sobolev(X, Y):
    """
    Sobolev kernel of degree 2.

    Parameters
    ----------
    X
        Samples of input variable
    Y
        Samples of output variable

    Returns
    -------
    Kernel matrix
    """
    XX = tf.math.abs(X - Y)
    B2XX = XX**2 - XX + 1 / 6

    Xr1 = tf.math.abs(X) - 0.5
    Xr2 = tf.math.abs(Y) - 0.5
    B1XX = tf.einsum("ijkl, ijlm-> ijkm", Xr1, Xr2)

    K = B2XX / 2 + B1XX
    return K


class Kernel(Enum):
    """
    GSA Perturbation function interface.
    """

    RBF = rbf
    BINARY = binary
    SOBOLEV = sobolev

    @staticmethod
    def from_string(kernel_function: str) -> "Kernel":
        """
        Restore a perturbation function from a string.

        Parameters
        ----------
        perturbation_function
            String indicating the perturbation function to restore: must be one
            of 'inpainting', 'blurring' or 'amplitude'.

        Returns
        -------
        perturbation_function
            The PerturbationFunction object.
        """
        assert kernel_function in [
            "rbf",
            "binary",
            "sobolev",
        ], "Only 'rbf', 'binary' and 'sobolev' are supported."

        if kernel_function == "rbf":
            return Kernel.RBF
        if kernel_function == "binary":
            return Kernel.BINARY
        return Kernel.SOBOLEV



import numpy as np

from xplique.attributions import HsicAttributionMethod
from xplique.attributions.global_sensitivity_analysis import (
    SobolevEstimator,
    BinaryEstimator,
    RbfEstimator,
)
from ..utils import generate_data, generate_model, almost_equal


def test_hsic_kernels_shape():
    """
    Test if the kernels are correctly computed.
    """
    input_shapes = [(20, 20, 1), (20, 20, 3)]
    nb_labels = 10

    for input_shape in input_shapes:
        x, y = generate_data(input_shape, nb_labels, 10)
        model = generate_model(input_shape, nb_labels)

        method = HsicAttributionMethod(model, grid_size=2, nb_design=8)
        stis_maps = method.explain(x, y)

        assert x.shape[:-1] == stis_maps.shape


def test_estimators():
    """Ensure every proposed estimator is working"""

    estimators = [SobolevEstimator(), BinaryEstimator(), RbfEstimator()]
    input_shapes = [(20, 20, 1), (20, 20, 3)]
    nb_labels = 10

    for input_shape in input_shapes:
        x, y = generate_data(input_shape, nb_labels, 10)
        model = generate_model(input_shape, nb_labels)

        for estimator in estimators:
            method = HsicAttributionMethod(
                model, grid_size=2, nb_design=8, estimator=estimator
            )
            stis_maps = method.explain(x, y)

            assert x.shape[:-1] == stis_maps.shape



"""
Global Sensitivity Analysis based Attribution Methods
"""
from .sobol_estimators import (
    JansenEstimator,
    JanonEstimator,
    HommaEstimator,
    SaltelliEstimator,
    GlenEstimator,
)
from .hsic_estimators import (
    HsicEstimator,
    BinaryEstimator,
    RbfEstimator,
    SobolevEstimator,
)
from .perturbations import inpainting, blurring, amplitude
from .samplers import (
    ScipySampler,
    TFSobolSequence,
    ScipySobolSequence,
    HaltonSequence,
    LatinHypercube,
)
from .replicated_designs import (
    LatinHypercubeRS,
    HaltonSequenceRS,
    ScipySobolSequenceRS,
    TFSobolSequenceRS,
)
from .sobol_attribution_method import SobolAttributionMethod
from .hsic_attribution_method import HsicAttributionMethod



import tensorflow as tf
import numpy as np

from xplique.commons import guided_relu_policy, deconv_relu_policy, override_relu_gradient, \
    open_relu_policy
from ..utils import almost_equal


def test_guided_policy():
    """Ensure the forward pass is the same and the guided mechanism is correct"""

    model1 = tf.keras.models.Sequential()
    model1.add(tf.keras.layers.Activation(tf.nn.relu, input_shape=(4,)))
    model1.add(tf.keras.layers.Lambda(lambda x: -x ** 2))
    model1_guided = override_relu_gradient(model1, guided_relu_policy)

    model2 = tf.keras.models.Sequential()
    model2.add(tf.keras.layers.ReLU(input_shape=(4,)))
    model2.add(tf.keras.layers.Lambda(lambda x: -x ** 2))
    model2_guided = override_relu_gradient(model2, guided_relu_policy)

    model3 = tf.keras.models.Sequential()
    model3.add(tf.keras.layers.ReLU(max_value=3, input_shape=(4,))) # relu with params
    model3.add(tf.keras.layers.Lambda(lambda x: -x ** 2))
    model3_guided = override_relu_gradient(model3, guided_relu_policy)

    x = tf.constant(np.expand_dims([5.0, 0.0, -5.0, 10.0], axis=0))
    y1, y2, y3 = model1(x), model2(x), model3(x)
    y1g, y2g, y3g = model1_guided(x), model2_guided(x), model3_guided(x)

    # the policy should not impact the forward pass
    assert almost_equal(y1, y1g)
    assert almost_equal(y2, y2g)
    assert almost_equal(y3, y3g)

    # gradient of model 1, 2 and 3 should be the same
    with tf.GradientTape() as tape:
        tape.watch(x)
        y1g = model1_guided(x)
    guided_grads_1 = tape.gradient(y1g, x).numpy()[0]

    with tf.GradientTape() as tape:
        tape.watch(x)
        y2g = model2_guided(x)
    guided_grads_2 = tape.gradient(y2g, x).numpy()[0]

    with tf.GradientTape() as tape:
        tape.watch(x)
        y3g = model3_guided(x)
    guided_grads_3 = tape.gradient(y3g, x).numpy()[0]

    assert almost_equal(guided_grads_1, guided_grads_2)
    assert almost_equal(guided_grads_1, guided_grads_3)
    assert almost_equal(guided_grads_1, np.array([0.0, 0.0, 0.0, 0.0]))


def test_deconv_policy():
    """Ensure the forward pass is the same and the guided mechanism is correct"""

    model1 = tf.keras.models.Sequential()
    model1.add(tf.keras.layers.Activation(tf.nn.relu, input_shape=(4,)))
    model1.add(tf.keras.layers.Lambda(lambda x: x ** 2))
    model1_deconv = override_relu_gradient(model1, deconv_relu_policy)

    model2 = tf.keras.models.Sequential()
    model2.add(tf.keras.layers.ReLU(input_shape=(4,)))
    model2.add(tf.keras.layers.Lambda(lambda x: x ** 2))
    model2_deconv = override_relu_gradient(model2, deconv_relu_policy)

    model3 = tf.keras.models.Sequential()
    model3.add(tf.keras.layers.ReLU(max_value=3, input_shape=(4,))) # relu with params
    model3.add(tf.keras.layers.Lambda(lambda x: x ** 2))
    model3_deconv = override_relu_gradient(model3, deconv_relu_policy)

    x = tf.constant(np.expand_dims([5.0, 0.0, -5.0, 10.0], axis=0))
    y1, y2, y3 = model1(x), model2(x), model3(x)
    y1d, y2d, y3d = model1_deconv(x), model2_deconv(x), model3_deconv(x)

    # the policy should not impact the forward pass
    assert almost_equal(y1, y1d)
    assert almost_equal(y2, y2d)
    assert almost_equal(y3, y3d)

    # gradient of model 1, 2 should be the same
    with tf.GradientTape() as tape:
        tape.watch(x)
        y1d = model1_deconv(x)
    deconv_grads_1 = tape.gradient(y1d, x).numpy()[0]

    with tf.GradientTape() as tape:
        tape.watch(x)
        y2d = model2_deconv(x)
    deconv_grads_2 = tape.gradient(y2d, x).numpy()[0]

    assert almost_equal(deconv_grads_1, deconv_grads_2)
    assert almost_equal(deconv_grads_1, np.array([10.0, 0.0, 0.0, 20.0]))


    with tf.GradientTape() as tape:
        tape.watch(x)
        y3d = model3_deconv(x)
    deconv_grads_3 = tape.gradient(y3d, x).numpy()[0]

    assert almost_equal(deconv_grads_3, np.array([6.0, 0.0, 0.0, 6.0]))


def test_open_relu():
    """Ensure the backward pass let all the gradient flow"""

    model1 = tf.keras.models.Sequential()
    model1.add(tf.keras.layers.Lambda(lambda x: x ** 3, input_shape=(4,)))
    model1.add(tf.keras.layers.Activation(tf.nn.relu))
    model1_open = override_relu_gradient(model1, open_relu_policy)

    model2 = tf.keras.models.Sequential()
    model2.add(tf.keras.layers.Lambda(lambda x: x ** 3, input_shape=(4,)))
    model2.add(tf.keras.layers.ReLU())
    model2_open = override_relu_gradient(model2, open_relu_policy)

    model3 = tf.keras.models.Sequential()
    model3.add(tf.keras.layers.Lambda(lambda x: x ** 3, input_shape=(4,)))
    model3.add(tf.keras.layers.ReLU(max_value=3,)) # relu with params
    model3_open = override_relu_gradient(model3, open_relu_policy)

    x = tf.constant(np.expand_dims([5.0, 0.0, -5.0, 10.0], axis=0))
    y1, y2, y3 = model1(x), model2(x), model3(x)
    y1d, y2d, y3d = model1_open(x), model2_open(x), model3_open(x)

    # the policy should not impact the forward pass
    assert almost_equal(y1, y1d)
    assert almost_equal(y2, y2d)
    assert almost_equal(y3, y3d)

    # gradient of model 1, 2 should be the same
    with tf.GradientTape() as tape:
        tape.watch(x)
        y1d = model1_open(x)
    open_grads_1 = tape.gradient(y1d, x).numpy()[0]

    with tf.GradientTape() as tape:
        tape.watch(x)
        y2d = model2_open(x)
    open_grads_2 = tape.gradient(y2d, x).numpy()[0]

    assert almost_equal(open_grads_1, open_grads_2)
    assert almost_equal(open_grads_1, 3.0 * x ** 2.0) # dx**3/dx = 3x**2


    with tf.GradientTape() as tape:
        tape.watch(x)
        y3d = model3_open(x)
    open_grads_3 = tape.gradient(y3d, x).numpy()[0]

    assert almost_equal(open_grads_3, 3.0 * x ** 2.0)



"""
Feature Visualization module
"""

from .losses import cosine_similarity
from .regularizers import l1_reg, l2_reg, total_variation_reg
from .transformations import random_blur, random_jitter, random_scale, \
                             random_flip, pad, compose_transformations
from .preconditioning import fft_image, get_fft_scale, fft_to_rgb, \
                             init_maco_buffer, maco_image_parametrization
from .objectives import Objective
from .optim import optimize
from .maco import maco



"""
Sobol Attribution Method explainer
"""

from ...types import Callable, Union, Optional, OperatorSignature
from ...commons import Tasks
from .gsa_attribution_method import GSABaseAttributionMethod
from .sobol_estimators import SobolEstimator, JansenEstimator
from .replicated_designs import ReplicatedSampler, TFSobolSequenceRS


class SobolAttributionMethod(GSABaseAttributionMethod):
    """
    Sobol' Attribution Method.
    Compute the total order Sobol' indices using a perturbation function on a grid and an
    adapted sampling as described in the original paper.

    Ref. Fel, Cadène, Chalvidal & al., Look at the Variance! Efficient Black-box Explanations
    with Sobol-based Sensitivity Analysis, NeurIPS (2021), https://arxiv.org/abs/2111.04138

    Parameters
    ----------
    model
        Model used for computing explanations.
    grid_size
        Cut the image in a grid of (grid_size, grid_size) to estimate an indice per cell.
    nb_design
        Must be a power of two. Number of design, the number of forward
        will be: nb_design * (grid_size**2 + 2). Generally not above 32.
    sampler
        Sampler used to generate the (quasi-)monte carlo samples, QMC (sobol sequence
        recommended). For more option, see the sampler module.
    estimator
        Estimator used to compute the total order sobol' indices, Jansen recommended. For more
        option, see the estimator module.
    perturbation_function
        Function to call to apply the perturbation on the input. Can also be string in
        'inpainting', 'blur'.
    batch_size
        Batch size to use for the forwards.
    operator
        Function g to explain, g take 3 parameters (f, x, y) and should return a scalar,
        with f the model, x the inputs and y the targets. If None, use the standard
        operator g(f, x, y) = f(x)[y].
    """

    def __init__(
        self,
        model,
        grid_size: int = 8,
        nb_design: int = 32,
        sampler: Optional[ReplicatedSampler] = None,
        estimator: Optional[SobolEstimator] = None,
        perturbation_function: Optional[Union[Callable, str]] = "inpainting",
        batch_size=256,
        operator: Optional[Union[Tasks, str, OperatorSignature]] = None,
    ):

        assert (
            nb_design & (nb_design - 1) == 0
        ) and nb_design != 0, "The number of design must be a power of two."

        sampler = sampler if sampler is not None else TFSobolSequenceRS()
        estimator = estimator if estimator is not None else JansenEstimator()

        assert isinstance(sampler, ReplicatedSampler), (
            "The sampler must be a" " valid Replicated Sampler."
        )
        assert isinstance(estimator, SobolEstimator), (
            "The estimator must be a" " valid Sobol estimator."
        )

        super().__init__(
            model=model,
            operator=operator,
            sampler=sampler,
            estimator=estimator,
            grid_size=grid_size,
            nb_design=nb_design,
            perturbation_function=perturbation_function,
            batch_size=batch_size,
        )



"""
Ensure we can use the operator functionnality on various models
"""

import numpy as np
import tensorflow as tf

import pytest

from xplique.attributions import (Saliency, GradientInput, IntegratedGradients, SmoothGrad, VarGrad,
                                  SquareGrad, GradCAM, Occlusion, Rise, GuidedBackprop, DeconvNet,
                                  GradCAMPP, Lime, KernelShap, SobolAttributionMethod,
                                  HsicAttributionMethod)
from xplique.commons.operators import (predictions_operator, regression_operator)
from ..utils import generate_data, generate_model, generate_regression_model, almost_equal


def default_methods(model, operator):
    return [
        Saliency(model, operator=operator),
        GradientInput(model, operator=operator),
        SmoothGrad(model, operator=operator),
        VarGrad(model, operator=operator),
        SquareGrad(model, operator=operator),
        IntegratedGradients(model, operator=operator),
        Occlusion(model, operator=operator),
        Rise(model, operator=operator, nb_samples=2),
        GuidedBackprop(model, operator=operator),
        DeconvNet(model, operator=operator),
        SobolAttributionMethod(model, operator=operator, grid_size=2, nb_design=2),
        HsicAttributionMethod(model, operator=operator, grid_size=2, nb_design=2),
    ]


def get_concept_model():
    model = tf.keras.Sequential([
        tf.keras.layers.Input((6)),
        tf.keras.layers.Dense((10))
    ])
    model.compile()
    return model


def test_predictions_operator():
    input_shape, nb_labels, samples = ((10, 10, 3), 10, 20)
    x, y = generate_data(input_shape, nb_labels, samples)
    classification_model = generate_model(input_shape, nb_labels)

    methods = default_methods(classification_model, regression_operator)
    for method in methods:

        assert hasattr(method, 'inference_function')
        assert hasattr(method, 'batch_inference_function')
        assert hasattr(method, 'gradient')
        assert hasattr(method, 'batch_gradient')

        phis = method(x, y)

        assert x.shape[:-1] == phis.shape[:3]



def test_regression_operator():
    input_shape, nb_labels, samples = ((10, 10, 1), 10, 20)
    x, y = generate_data(input_shape, nb_labels, samples)
    regression_model = generate_regression_model(input_shape, nb_labels)

    methods = default_methods(regression_model, regression_operator)
    for method in methods:

        assert hasattr(method, 'inference_function')
        assert hasattr(method, 'batch_inference_function')
        assert hasattr(method, 'gradient')
        assert hasattr(method, 'batch_gradient')

        phis = method(x, y)

        assert x.shape[:-1] == phis.shape[:3]


def test_concept_operator():
    concept_model = get_concept_model()

    x, y = generate_data((20, 20, 1), 10, 10)

    random_projection = tf.cast(np.random.uniform(size=(20*20, 6)), tf.float32)

    def concept_operator(model, x, y):
        x = tf.reshape(x, (-1, 20*20))
        ui = x @ random_projection
        return tf.reduce_sum(model(ui) * y, axis=-1)

    methods = default_methods(concept_model, concept_operator)

    for method in methods:

        assert hasattr(method, 'inference_function')
        assert hasattr(method, 'batch_inference_function')
        assert hasattr(method, 'gradient')
        assert hasattr(method, 'batch_gradient')

        phis = method(x, y)

        assert x.shape[:-1] == phis.shape[:3]



import numpy as np

from xplique.attributions import SobolAttributionMethod
from xplique.attributions.global_sensitivity_analysis import (LatinHypercubeRS,
    ScipySobolSequenceRS, TFSobolSequenceRS, HaltonSequenceRS, JansenEstimator,
    JanonEstimator, GlenEstimator, SaltelliEstimator,HommaEstimator, inpainting,
    blurring, amplitude)
from ..utils import generate_data, generate_model, almost_equal


def test_output_shape():
    """The output size (h, w) must be the same as the input"""

    input_shapes = [(20, 20, 1), (20, 20, 3)]
    nb_labels = 10

    for input_shape in input_shapes:
        x, y = generate_data(input_shape, nb_labels, 10)
        model = generate_model(input_shape, nb_labels)

        method = SobolAttributionMethod(model,grid_size = 2, nb_design = 8)
        stis_maps = method.explain(x, y)

        assert x.shape[:-1] == stis_maps.shape


def test_samplers():
    """Ensure every proposed sampler is working"""

    samplers_class = [LatinHypercubeRS, ScipySobolSequenceRS, TFSobolSequenceRS, HaltonSequenceRS]
    input_shapes = [(20, 20, 1), (20, 20, 3)]
    nb_labels = 10


    for input_shape in input_shapes:

        x, y = generate_data(input_shape, nb_labels, 10)
        model = generate_model(input_shape, nb_labels)

        for sampler_c in samplers_class:
            # (fel) some sampler(s) need scipy>=1.7, we should enforce it at some point (py>=3.6)
            try:
                sampler = sampler_c()
                method = SobolAttributionMethod(model, grid_size = 2, nb_design = 8, sampler = sampler)
                stis_maps = method.explain(x, y)

                assert x.shape[:-1] == stis_maps.shape
            except ModuleNotFoundError:
                pass


def test_perturbations():
    """Ensure every proposed perturbations is working"""

    perturbations = [inpainting, blurring, amplitude]
    input_shapes = [(20, 20, 1), (20, 20, 3)]
    nb_labels = 10

    for input_shape in input_shapes:
        x, y = generate_data(input_shape, nb_labels, 10)
        model = generate_model(input_shape, nb_labels)

        for perturb in perturbations:
            method = SobolAttributionMethod(model, grid_size = 2, nb_design = 8, perturbation_function = perturb)
            stis_maps = method.explain(x, y)

            assert x.shape[:-1] == stis_maps.shape


def test_estimators():
    """Ensure every proposed estimator is working"""

    estimators = [JanonEstimator(), JansenEstimator(), GlenEstimator(), SaltelliEstimator(), HommaEstimator()]
    input_shapes = [(20, 20, 1), (20, 20, 3)]
    nb_labels = 10

    for input_shape in input_shapes:
        x, y = generate_data(input_shape, nb_labels, 10)
        model = generate_model(input_shape, nb_labels)

        for estimator in estimators:
            method = SobolAttributionMethod(model, grid_size = 2, nb_design = 8, estimator = estimator)
            stis_maps = method.explain(x, y)

            assert x.shape[:-1] == stis_maps.shape


def test_ishigami():
    """Ensure every proposed estimator compute the correct Sobol total indice estimator on the ishigami function"""

    # Ishigami function
    # Y = sin(X1) + a sin(X2)^2 + b * X3^4 * sin(X1)
    # X in U([-pi, pi])^3

    # analytically we find:
    # V(Y) = a^2 / 8 + b*pi^4/5 + b^2pi^8/18 + 1/2

    # V1 = b * pi^4 / 5 + b^2 * pi^8 / 50 + 1/2
    # V2 = a^2/8
    # V3 = 0

    # V12 = 0, V23 = 0, V13 = 8b^2pi^8/225
    # V13 = 8b^2 pi^8 / 255
    # V23 = 0
    # V123 = 0

    # S1 = V1 / V(Y)
    # S2 = V2 / V(Y)
    # S3 = V3 / V(Y)
    # S13 = V13 / V(Y)

    # ST1 = S1 + S12 + S13 + S123 = S1 + S13
    # ST2 = S2 + S12 + S23 + S123 = S2
    # ST3 = S3 + S13 + S23 + S123 = S13

    def ishigami_function(X, a, b):
        return np.sin(X[:, 0]) + a * np.sin(X[:, 1])**2.0 + b * np.sin(X[:, 0]) * X[:, 2]**4.0

    n = 2**15
    a = 7
    b = 0.1

    # order 1
    Vy = a**2.0 / 8.0 + b * np.pi**4.0 / 5.0 + b**2.0 * np.pi**8.0 / 18.0 + 0.5
    V1 = b * np.pi**4.0 / 5 + b**2.0 * np.pi**8.0 / 50.0 + 0.5
    V2 = a**2.0 / 8.0
    V3 = 0.0

    # order 2
    V12 = 0.0
    V13 = 8.0 * b**2.0 * np.pi**8.0 / 225.0
    V23 = 0.0

    # order 3
    V123 = 0.0

    S1 = V1 / Vy
    S2 = V2 / Vy
    S3 = V3 / Vy

    S12 = V12 / Vy
    S13 = V13 / Vy
    S23 = V23 / Vy

    S123 = V123 / Vy

    ST1 = S1 + S12 + S13 + S123
    ST2 = S2 + S12 + S23 + S123
    ST3 = S3 + S13 + S23 + S123

    STis = np.array([ST1, ST2, ST3])

    X = TFSobolSequenceRS()(3, n)
    X *= 2.0 * np.pi
    X -= np.pi

    Y = ishigami_function(X, a, b)

    estimators = [JanonEstimator(), JansenEstimator(), GlenEstimator(), SaltelliEstimator(), HommaEstimator()]

    for est in estimators:
        predicted_STis = est(X, Y, n)
        assert almost_equal(STis, predicted_STis, 1e-3)






import numpy as np

from xplique.features_visualizations.losses import cosine_similarity
from ..utils import almost_equal


def test_cosine_similarity():
    vec = np.array([10.0, 20.0, 30.0])[np.newaxis, :]
    vec_colinear = np.array([1.0, 2.0, 3.0])[np.newaxis, :]
    vec_orthogonal = np.array([.0, .0, .0])[np.newaxis, :]
    vec_opposite = np.array([-0.01, -0.02, -.03])[np.newaxis, :]

    # cosine_similarity(a, b) = <a,b> / (|a| + |b|)
    assert almost_equal(cosine_similarity(vec, vec_colinear)[0],   1.0)
    assert almost_equal(cosine_similarity(vec, vec_orthogonal)[0], .0)
    assert almost_equal(cosine_similarity(vec, vec_opposite)[0],  -1.0)



"""
Operator for object detection
"""
from typing import Tuple
import tensorflow as tf

_EPSILON = tf.constant(1e-4)


def _box_iou(boxes_a: tf.Tensor, boxes_b: tf.Tensor) -> tf.Tensor:
    """
    Compute the intersection between two batched bounding boxes.
    Each bounding box is defined by (x1, y1, x2, y2) respectively (left, bottom, right, top).
    With left < right and bottom < top

    Parameters
    ----------
    boxes_a
        First batch of bounding boxes.
    boxes_b
        Second batch of bounding boxes.

    Returns
    -------
    iou_score
        The IOU score between the two batches of bounding boxes.
    """

    # determine the intersection rectangle
    left   = tf.maximum(boxes_a[..., 0], boxes_b[..., 0])
    bottom = tf.maximum(boxes_a[..., 1], boxes_b[..., 1])
    right  = tf.minimum(boxes_a[..., 2], boxes_b[..., 2])
    top    = tf.minimum(boxes_a[..., 3], boxes_b[..., 3])

    intersection_area = tf.math.maximum(right - left, 0) * tf.math.maximum(top - bottom, 0)

    # determine the areas of the prediction and ground-truth rectangles
    a_area = (boxes_a[..., 2] - boxes_a[..., 0]) * (boxes_a[..., 3] - boxes_a[..., 1])
    b_area = (boxes_b[..., 2] - boxes_b[..., 0]) * (boxes_b[..., 3] - boxes_b[..., 1])

    union_area = a_area + b_area - intersection_area

    iou_score = intersection_area / (union_area + _EPSILON)

    return iou_score


def _format_objects(predictions: tf.Tensor) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
    """
    Format bounding boxes prediction for object detection operator.
    It takes a batch of bounding boxes predictions of the model and divide it between
    boxes_coordinates, proba_detection, and one_hots_classifications.

    Parameters
    ----------
    predictions
        Batch of bounding boxes predictions of shape (nb_boxes, (4 + 1 + nc)).
        (4 + 1 + nc) means: [boxes_coordinates, proba_detection, one_hots_classifications].
        Where nc is the number of classes.

    Returns
    -------
    boxes_coordinates
        A Tensor of shape (nb_boxes, 4) encoding the boxes coordinates.
    proba_detection
        A Tensor of shape (nb_boxes, 1) encoding the detection probabilities.
    one_hots_classifications
        A Tensor of shape (nb_boxes, nc) encoding the class predictions.
    """
    boxes_coordinates, proba_detection, one_hots_classifications = \
        tf.split(predictions, [4, 1, tf.shape(predictions[0])[0] - 5], 1)
    return boxes_coordinates, proba_detection, one_hots_classifications



import tensorflow as tf
import numpy as np

from ..utils import generate_model, generate_data, almost_equal
from xplique.metrics import AverageStability


def test_average_stability():
    # ensure we can compute the metric with consistents arguments
    input_shape, nb_labels, nb_samples = ((32, 32, 3), 10, 20)
    x, y = generate_data(input_shape, nb_labels, nb_samples)
    model = generate_model(input_shape, nb_labels)
    explainer = lambda x, y : np.random.uniform(0, 1, x.shape[:-1])

    l_inf_dist = lambda phi_a, phi_b: np.max(phi_a - phi_b)

    for batch_size in [64, None]:
        for dist in ['l1', 'l2', l_inf_dist]:
            score = AverageStability(model, x, y,
                                     batch_size=batch_size,
                                     radius=0.1,
                                     distance=dist,
                                     nb_samples=100)(explainer)

            assert isinstance(score, float)
            assert score < np.prod(x.shape[:-1])


def test_noise_radius():
    # ensure the noise added correspond to the radius passed.
    input_shape, nb_labels, nb_samples = ((32, 32, 3), 10, 20)
    x, y = generate_data(input_shape, nb_labels, nb_samples)
    model = generate_model(input_shape, nb_labels)

    mean_noise = []

    for radius in [1.0 * 10**i for i in range(-4, 4)]:
        metric = AverageStability(model, x, y, radius=radius, nb_samples=1000)
        mean_noise.append(np.mean(metric.noisy_masks))

    # the list should striclty increase
    assert all(i < j for i, j in zip(mean_noise, mean_noise[1:]))


def test_perfect_stability():
    """Ensure we get perfect stability when explanation is the same"""
    input_shape, nb_labels, nb_samples = ((8, 8, 1), 2, 20)
    x, y = generate_data(input_shape, nb_labels, nb_samples)
    model = generate_model(input_shape, nb_labels)

    explainer = lambda x, y: tf.ones(x.shape)

    perfect_score = AverageStability(model, x, y)(explainer)
    assert almost_equal(perfect_score, 0.0)


def test_worst_stability():
    """Ensure we get worst stability when explanation is totally different"""
    dim = 8
    input_shape, nb_labels, nb_samples = ((dim, dim, 1), 2, 20)
    x, y = generate_data(input_shape, nb_labels, nb_samples)
    model = model = generate_model(input_shape, nb_labels)

    def worst_explainer(inputs, _):
        is_original = any([np.sum(np.abs(inputs[0] - _x)) == .0 for _x in x])
        if is_original:
            return np.ones(inputs.shape)
        else:
            return np.zeros(inputs.shape)

    worst_score = AverageStability(model, x, y)(worst_explainer)
    assert almost_equal(worst_score, dim**2)



"""
Functions to prepare `targets` for segmentation attributions
"""

import cv2
import numpy as np
import tensorflow as tf

from ..types import Union, Tuple, List


def get_class_zone(predictions: Union[tf.Tensor, np.array], class_id: int) -> tf.Tensor:
    """
    Extract a mask for the class `c`.
    The mask correspond to the pixels where the maximum prediction correspond to the class `c`.
    Other classes channels are set to zero.

    Parameters
    ----------
    predictions
        Output of the model, it should be the output of a softmax function.
        We assume the shape (h, w, c).
    class_id
        Index of the channel of the class of interest.

    Returns
    -------
    class_zone_mask
        Mask of the zone corresponding to the class of interest.
        Only the corresponding channel is non-zero.
        The shape is the same as `predictions`, (h, w, c).
    """
    assert len(tf.shape(predictions)) == 3, "predictions should correspond to only one image"

    class_zone = tf.cast(tf.argmax(predictions, axis=-1) == class_id, tf.float32)
    class_zone_mask = tf.Variable(tf.zeros(predictions.shape))
    class_zone_mask = class_zone_mask[:, :, class_id].assign(class_zone)

    assert tf.reduce_sum(class_zone_mask) >= 1

    return class_zone_mask


def get_connected_zone(
    predictions: Union[tf.Tensor, np.array], coordinates: Tuple[int, int]
) -> tf.Tensor:
    """
    Extract a connected mask around `coordinates`.
    The mask correspond to the pixels where the maximum prediction correspond
    to the maximum predicted class at `coordinates`.
    This class mask is then limited to the connected zone around `coordinates`.
    Other classes channels are set to zero.

    Parameters
    ----------
    predictions
        Output of the model, it should be the output of a softmax function.
        We assume the shape (h, w, c).
    coordinates
        Tuple of coordinates of the point inside the zone of interest.

    Returns
    -------
    connected_zone_mask
        Mask of the connected zone around `coordinates` with similar class prediction.
        Only the corresponding channel is non-zero.
        The shape is the same as `predictions`, (h, w, c).
    """
    assert len(tf.shape(predictions)) == 3, "predictions should correspond to only one image"

    assert (
        coordinates[0] < predictions.shape[0]
    ), f"Coordinates should be included in the shape, i.e. {coordinates[0]}<{predictions.shape[0]}"
    assert (
        coordinates[1] < predictions.shape[1]
    ), f"Coordinates should be included in the shape, i.e. {coordinates[1]}<{predictions.shape[1]}"

    labels = tf.argmax(predictions, axis=-1)
    class_id = labels[coordinates[0], coordinates[1]]
    mask = labels == class_id
    mask = np.uint8(np.array(mask)[:, :, np.newaxis] * 255)

    components_masks = cv2.connectedComponents(mask)[1]  # pylint: disable=no-member

    component_id = components_masks[coordinates[0], coordinates[1]]
    connected_zone = tf.cast(components_masks == component_id, tf.float32)

    connected_zone_mask = tf.Variable(tf.zeros(predictions.shape))
    connected_zone_mask = connected_zone_mask[:, :, class_id].assign(connected_zone)

    assert tf.reduce_sum(connected_zone_mask) >= 1
    assert connected_zone_mask[coordinates[0], coordinates[1], class_id] != 0

    return connected_zone_mask


def list_class_connected_zones(
        predictions: Union[tf.Tensor,np.array],
        class_id: int,
        zone_minimum_size: int = 100
) -> List[tf.Tensor]:
    """
    List all connected zones for a given class.
    A connected zone is a set of pixels next to each others
    where the maximum prediction correspond to the same class.
    This function generate a list of connected zones,
    each element of the list have a similar format to `get_connected_zone` outputs.

    Parameters
    ----------
    predictions
        Output of the model, it should be the output of a softmax function.
        We assume the shape (h, w, c).
    class_id
        Index of the channel of the class of interest.
    zone_minimum_size
        Threshold of number of pixels under which zones are not returned.

    Returns
    -------
    connected_zones_masks_list
        List of the connected zones masks for a given class.
        Each zone predictions shape is the same as `predictions`, (h, w, c).
        Only the corresponding channel is non-zero.
    """
    assert len(tf.shape(predictions)) == 3, "predictions should correspond to only one image"

    labels = tf.argmax(predictions, axis=-1)
    mask = labels == class_id
    mask = np.uint8(np.array(mask)[:, : , np.newaxis] * 255)

    components_masks = cv2.connectedComponents(mask)[1]  # pylint: disable=no-member

    sizes = np.bincount(components_masks.ravel())

    connected_zones_masks_list = []
    for component_id, size in enumerate(sizes[1:]):

        if size > zone_minimum_size:

            connected_zone = tf.cast(components_masks == (component_id + 1), tf.float32)

            all_channels_class_zone_mask = tf.Variable(tf.zeros(predictions.shape))
            all_channels_class_zone_mask =\
                all_channels_class_zone_mask[:, :, class_id].assign(connected_zone)

            assert tf.reduce_sum(all_channels_class_zone_mask) >= 1

            connected_zones_masks_list.append(all_channels_class_zone_mask)

    return connected_zones_masks_list



def get_in_out_border(
    class_zone_mask: Union[tf.Tensor, np.array],
) -> tf.Tensor:
    """
    Extract the border of a zone of interest, then put `1` on the
    inside border and `-1` on the outside border.

    Examples of coefficients extracted from the class channel of the class of interest:

    ```
    # class_zone_mask[:, :, c]
    [[0, 0, 0, 0, 0],
     [0, 0, 0, 1, 1],
     [0, 0, 1, 1, 1],
     [0, 0, 1, 1, 1],
     [0, 1, 1, 1, 1]]

    # border_mask
    [[ 0,  0, -1, -1, -1],
     [ 0, -1, -1,  1,  1],
     [ 0, -1,  1,  1,  0],
     [-1, -1,  1,  0,  0],
     [-1,  1,  1,  0,  0]]
    ```

    Parameters
    ----------
    class_zone_mask
        Mask delimiting the zone of interest,  for the class of interest
        only one channel should have non-zero values,
        the one corresponding to the class.
        We assume the shape (h, w, c) same as the model output for one element.

    Returns
    -------
    class_borders_masks
        Mask of the borders of the zone of the class of interest.
        Only the corresponding channel is non-zero.
        Inside borders are set to `1` and outside borders are set to `-1`.
        The shape is the same as `class_zone_mask`, (h, w, c).
    """
    assert len(tf.shape(class_zone_mask)) == 3,\
        "class_zone_mask should correspond to only one image"

    # channel of the class of interest
    channel_mean = tf.reduce_sum(tf.cast(class_zone_mask, tf.int32), axis=[0, 1])
    assert (
        int(tf.reduce_sum(channel_mean)) >= 1
    ), "The specified `class_target_mask` is empty."
    class_id = int(tf.argmax(channel_mean))

    # set other values to -1 on the target zone to 1
    binary_mask = 2 * tf.cast(class_zone_mask[:, :, class_id], tf.int32) - 1

    # extend size with padding for convolution
    extended_binary_mask = tf.pad(
        binary_mask,
        tf.constant([[1, 1], [1, 1]]),
        "SYMMETRIC",
    )

    kernel = tf.convert_to_tensor(
        [[-1, -1, -1], [-1, 13, -1], [-1, -1, -1]], dtype=tf.int32
    )
    kernel = kernel[:, :, tf.newaxis, tf.newaxis]
    conv_result = tf.nn.conv2d(
        tf.expand_dims(tf.expand_dims(extended_binary_mask, axis=0), axis=-1),
        kernel,
        strides=1,
        padding="VALID",
    )[0, :, :, 0]

    # 6 < in < 21, -21 < out < -6
    in_border = tf.logical_and(
        tf.math.less(tf.constant([6]), conv_result),
        tf.math.less(conv_result, tf.constant([21])),
    )
    out_border = tf.logical_and(
        tf.math.less(tf.constant([-21]), conv_result),
        tf.math.less(conv_result, tf.constant([-6])),
    )

    border_mask = (
        tf.zeros(binary_mask.shape)
        + tf.cast(in_border, tf.float32)
        - tf.cast(out_border, tf.float32)
    )

    class_borders_masks = tf.Variable(tf.zeros(class_zone_mask.shape))
    class_borders_masks = class_borders_masks[:, :, class_id].assign(border_mask)

    assert int(tf.reduce_sum(tf.abs(class_borders_masks))) >= 1

    return class_borders_masks


def get_common_border(
    border_mask_1: Union[tf.Tensor, np.array], border_mask_2: Union[tf.Tensor, np.array]
) -> tf.Tensor:
    """
    Compute the common part between `border_mask_1` and `border_mask_2` masks.
    Those borders should be computed using `get_in_out_border`.

    Parameters
    ----------
    border_mask_1
        Border of the first zone of interest. Computed with `get_in_out_border`.
    border_mask_2
        Border of the second zone of interest. Computed with `get_in_out_border`.

    Returns
    -------
    common_borders_masks
        Mask of the common borders between two zones of interest.
        Only the two corresponding channels are non-zero.
        Inside borders are set to `1` and outside borders are set to `-1`,
        Respectively on the two channels.
        The shape is the same as the input border masks, (h, w, c).
    """
    all_channel_border_mask_1 = tf.reduce_any(border_mask_1 != 0, axis=-1)
    all_channel_border_mask_2 = tf.reduce_any(border_mask_2 != 0, axis=-1)

    common_pixels_mask = tf.logical_and(
        all_channel_border_mask_1, all_channel_border_mask_2
    )

    assert tf.reduce_any(common_pixels_mask), "No common border between the two masks."

    return (border_mask_1 + border_mask_2) * tf.expand_dims(
        tf.cast(common_pixels_mask, tf.float32), -1
    )



"""
Module related to Occlusion sensitivity method
"""

from math import ceil

import numpy as np
import tensorflow as tf

from .base import BlackBoxExplainer, sanitize_input_output
from ..commons import repeat_labels, batch_tensor, Tasks
from ..types import Callable, Tuple, Union, Optional, OperatorSignature


class Occlusion(BlackBoxExplainer):
    """
    Used to compute the Occlusion sensitivity method, sweep a patch that occludes pixels over the
    images and use the variations of the model prediction to deduce critical areas.

    Ref. Zeiler & al., Visualizing and Understanding Convolutional Networks (2013).
    https://arxiv.org/abs/1311.2901

    Parameters
    ----------
    model
        The model from which we want to obtain explanations
    batch_size
        Number of pertubed samples to explain at once.
        Default to 32.
    operator
        Function g to explain, g take 3 parameters (f, x, y) and should return a scalar,
        with f the model, x the inputs and y the targets. If None, use the standard
        operator g(f, x, y) = f(x)[y].
    patch_size
        Size of the patches to apply, if integer then assume an hypercube.
    patch_stride
        Stride between two patches, if integer then assume an hypercube.
    occlusion_value
        Value used as occlusion.
    """

    def __init__(self,
                 model: Callable,
                 batch_size: Optional[int] = 32,
                 operator: Optional[Union[Tasks, str, OperatorSignature]] = None,
                 patch_size: Union[int, Tuple[int, int]] = 3,
                 patch_stride: Union[int, Tuple[int, int]] = 3,
                 occlusion_value: float = 0.0):
        super().__init__(model, batch_size, operator)

        self.patch_size = patch_size
        self.patch_stride = patch_stride
        self.occlusion_value = occlusion_value

    @sanitize_input_output
    def explain(self,
                inputs: Union[tf.data.Dataset, tf.Tensor, np.ndarray],
                targets: Optional[Union[tf.Tensor, np.ndarray]] = None) -> tf.Tensor:
        """
        Compute Occlusion sensitivity for a batch of samples.

        Parameters
        ----------
        inputs
            Dataset, Tensor or Array. Input samples to be explained.
            If Dataset, targets should not be provided (included in Dataset).
            Expected shape among (N, W), (N, T, W), (N, H, W, C).
            More information in the documentation.
        targets
            Tensor or Array. One-hot encoding of the model's output from which an explanation
            is desired. One encoding per input and only one output at a time. Therefore,
            the expected shape is (N, output_size).
            More information in the documentation.

        Returns
        -------
        explanations
            Occlusion sensitivity, same shape as the inputs, except for the channels.
        """

        # check if data is images
        is_image = len(inputs.shape) > 2

        if is_image:
            if not isinstance(self.patch_size, tuple):
                self.patch_size = (self.patch_size, self.patch_size)
            if not isinstance(self.patch_stride, tuple):
                self.patch_stride = (self.patch_stride, self.patch_stride)

        occlusion_maps = None
        batch_size = self.batch_size or len(inputs)

        masks = Occlusion._get_masks((*inputs.shape[1:],), self.patch_size, self.patch_stride)
        base_scores = self.batch_inference_function(self.model, inputs, targets, batch_size)

        # since the number of masks is often very large, we process the entries one by one
        for single_input, single_target, single_base_score in zip(inputs, targets, base_scores):

            occlusion_map = tf.zeros(masks.shape[1:])

            for batch_masks in batch_tensor(masks, batch_size):

                occluded_inputs = Occlusion._apply_masks(single_input, batch_masks,
                                                         self.occlusion_value)
                repeated_targets = repeat_labels(single_target[tf.newaxis, :], len(batch_masks))

                batch_scores = self.inference_function(self.model,
                                                       occluded_inputs,
                                                       repeated_targets)

                batch_sensitivity = Occlusion._compute_sensitivity(
                    single_base_score, batch_scores, batch_masks
                )

                occlusion_map += batch_sensitivity

            occlusion_maps = occlusion_map if occlusion_maps is None else \
                tf.concat([occlusion_maps, occlusion_map], axis=0)

        return occlusion_maps

    @staticmethod
    def _get_masks(input_shape: Union[Tuple[int, int, int], Tuple[int, int], Tuple[int]],
                   patch_size: Union[int, Tuple[int, int]],
                   patch_stride: Union[int, Tuple[int, int]]) -> tf.Tensor:
        """
        Create all the possible patches for the given configuration.

        Parameters
        ----------
        input_shape
            Desired shape, dimension of one sample.
        patch_size
            Size of the patches to apply.
        patch_stride
            Stride between two patches.

        Returns
        -------
        occlusion_masks
            The boolean occlusion masks, same shape as the inputs, with 1 as occluded.
        """
        masks = []

        # check if we have tabular data
        is_tabular = len(input_shape) == 1

        if is_tabular:
            x_anchors = [x * patch_stride for x in
                         range(0, ceil((input_shape[0] - patch_size + 1) / patch_stride))]

            for x_anchor in x_anchors:
                mask = np.zeros(input_shape, dtype=bool)
                mask[x_anchor:x_anchor + patch_size] = 1
                masks.append(mask)
        else:
            x_anchors = [x * patch_stride[0] for x in
                         range(0, ceil((input_shape[0] - patch_size[0] + 1) / patch_stride[0]))]
            y_anchors = [y * patch_stride[1] for y in
                         range(0, ceil((input_shape[1] - patch_size[1] + 1) / patch_stride[1]))]

            for x_anchor in x_anchors:
                for y_anchor in y_anchors:
                    mask = np.zeros(input_shape[:2], dtype=bool)
                    mask[x_anchor:x_anchor + patch_size[0], y_anchor:y_anchor + patch_size[1]] = 1
                    masks.append(mask)

        return tf.cast(masks, dtype=tf.bool)

    @staticmethod
    @tf.function
    def _apply_masks(current_input: tf.Tensor,
                     masks: tf.Tensor,
                     occlusion_value: float) -> tf.Tensor:
        """
        Given input samples and an occlusion mask template, apply it for every sample.

        Parameters
        ----------
        current_input
            Input samples to be explained.
        masks
            The boolean occlusion masks, with 1 as occluded.
        occlusion_value
            Value used as occlusion.

        Returns
        -------
        occluded_inputs
            All the occluded combinations for each inputs.
        """
        occluded_inputs = tf.expand_dims(current_input, axis=0)
        occluded_inputs = tf.repeat(occluded_inputs, repeats=masks.shape[0], axis=0)

        # check if current input shape is (W, H, C)
        has_channels = len(current_input.shape) > 2
        if has_channels:
            masks = tf.expand_dims(masks, axis=-1)
            masks = tf.repeat(masks, repeats=current_input.shape[-1], axis=-1)

        occluded_inputs = occluded_inputs * tf.cast(tf.logical_not(masks), tf.float32)
        occluded_inputs += tf.cast(masks, tf.float32) * occlusion_value

        return occluded_inputs

    @staticmethod
    @tf.function
    def _compute_sensitivity(baseline_scores: tf.Tensor,
                             occluded_scores: tf.Tensor,
                             masks: tf.Tensor) -> tf.Tensor:
        """
        Compute the sensitivity score given the score of the occluded inputs

        Parameters
        ----------
        baseline_scores
            Scores obtained with the original inputs (not occluded)
        occluded_scores
            Scores of the occluded combinations for the class of
            interest.
        masks
            The boolean occlusion masks, with 1 as occluded.

        Returns
        -------
        sensitivity
            Value reflecting the effect of each occlusion patchs on the output
        """
        baseline_scores = tf.expand_dims(baseline_scores, axis=-1)
        occluded_scores = tf.reshape(occluded_scores, (-1, masks.shape[0]))

        score_delta = baseline_scores - occluded_scores
        # reshape the delta score to fit masks
        score_delta = tf.reshape(score_delta, (*score_delta.shape, *(1,) * len(masks.shape[1:])))

        sensitivity = score_delta * tf.cast(masks, tf.float32)
        sensitivity = tf.reduce_sum(sensitivity, axis=1)

        return sensitivity
